/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __DBUS_H__
#define __DBUS_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.xfce.Dict */

#define TYPE_DICT (dict_get_type ())
#define DICT(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_DICT, Dict))
#define IS_DICT(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_DICT))
#define DICT_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), TYPE_DICT, DictIface))

struct _Dict;
typedef struct _Dict Dict;
typedef struct _DictIface DictIface;

struct _DictIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_search) (
    Dict *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_phrase);

};

GType dict_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *dict_interface_info (void);
guint dict_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void dict_complete_search (
    Dict *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void dict_call_search (
    Dict *proxy,
    const gchar *arg_phrase,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean dict_call_search_finish (
    Dict *proxy,
    GAsyncResult *res,
    GError **error);

gboolean dict_call_search_sync (
    Dict *proxy,
    const gchar *arg_phrase,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define TYPE_DICT_PROXY (dict_proxy_get_type ())
#define DICT_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_DICT_PROXY, DictProxy))
#define DICT_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), TYPE_DICT_PROXY, DictProxyClass))
#define DICT_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), TYPE_DICT_PROXY, DictProxyClass))
#define IS_DICT_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_DICT_PROXY))
#define IS_DICT_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), TYPE_DICT_PROXY))

typedef struct _DictProxy DictProxy;
typedef struct _DictProxyClass DictProxyClass;
typedef struct _DictProxyPrivate DictProxyPrivate;

struct _DictProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  DictProxyPrivate *priv;
};

struct _DictProxyClass
{
  GDBusProxyClass parent_class;
};

GType dict_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (DictProxy, g_object_unref)
#endif

void dict_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
Dict *dict_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
Dict *dict_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void dict_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
Dict *dict_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
Dict *dict_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define TYPE_DICT_SKELETON (dict_skeleton_get_type ())
#define DICT_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), TYPE_DICT_SKELETON, DictSkeleton))
#define DICT_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), TYPE_DICT_SKELETON, DictSkeletonClass))
#define DICT_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), TYPE_DICT_SKELETON, DictSkeletonClass))
#define IS_DICT_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), TYPE_DICT_SKELETON))
#define IS_DICT_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), TYPE_DICT_SKELETON))

typedef struct _DictSkeleton DictSkeleton;
typedef struct _DictSkeletonClass DictSkeletonClass;
typedef struct _DictSkeletonPrivate DictSkeletonPrivate;

struct _DictSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  DictSkeletonPrivate *priv;
};

struct _DictSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType dict_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (DictSkeleton, g_object_unref)
#endif

Dict *dict_skeleton_new (void);


G_END_DECLS

#endif /* __DBUS_H__ */
